# language=Markdown
"""
Builtins Wrapper -- typed overloads for the Python builtins methods.

## Builtins Wrapper

This module provides typed overloads for some Python builtins methods.
It contains no actual members -- all of them are shadowy wrapped to their builtins implementations.
"""

import builtins
from typing import *

T = TypeVar('T')
R = TypeVar('R')
def shadow_wrapper(other_func: Callable):
    """ Decoration for shadowy replace one function with the other """
    
    def wrapper(_: Callable):
        return other_func
    return wrapper

# noinspection PyShadowingBuiltins
@overload
def map(func: Callable[[T], R], coll: Iterable[T]) -> Iterator[R]:
    pass
# noinspection PyShadowingBuiltins
@shadow_wrapper(builtins.map)
def map(*args, **kwargs):
    return builtins.map(*args, **kwargs)


# noinspection PyShadowingBuiltins
@overload
def sum(iterable: Iterable[T]) -> T:
    pass
# noinspection PyShadowingBuiltins
@overload
def sum(iterable: Iterable[T], start: T) -> T:
    pass
# noinspection PyShadowingBuiltins
@shadow_wrapper(builtins.sum)
def sum(*args, **kwargs):
    return builtins.sum(*args, **kwargs)


__all__ = \
[
    'map',
    'sum',
]



# language=Markdown
"""
ChainTools module -- operations for method chaining and iterables mapping.

## ChainTools Module

This module provides functions for prettier method chaining.
Most of them are implemented by using `functools.reduce()` function.

### Provides:
#### Functions for iterating:
 * `chunks()`
 
#### Functions for mapping and applying:
 * `map_items()`
 * `apply()`
 * `apply_items()`
 
#### Functions with inversed arguments order:
 * `invcall()`
 * `invmap()`
 * `invmap_items()`
 
#### Functions for chaining:
 * `chain()`
 * `chain_map()`
 * `chain_map_items()`

Chain functions also provide typing overloads for up to 6 functions for the better type hinting.
"""

from functools import reduce, partial
from itertools import islice
from typing import *

from .predef import *

T  = TypeVar('T')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
T6 = TypeVar('T6')
R  = TypeVar('R')


def chunks(coll: Iterable[T], chunk_size: int) -> Iterator[List[T]]:
    """
    Splits an iterable into chunks of given size.
    We have no idea why this method is not part of `itertools` Python module.
    
    Internally implemented using `itertools.slice()`.
    
    Examples:
        ```python
        c = chunks(range(15), 6)
        print(list(c))
        # [[0, 1, 2, 3, 4, 5], [6, 7, 8, 9, 10, 11], [12, 13, 14]]
        ```
    
    Args:
        coll: `Iterable[T]`. An iterable to be split.
        chunk_size: `int`. A size of all but last chunks in the result. Last chunk could contain fewer number of elements.
    
    Returns:
        `Iterator[List[T]]`.
        Returns an iterator of lists made from the original iterable.
    """
    
    it = iter(coll)
    while True:
        head = list(islice(it, chunk_size))
        if (not head): break
        yield head

def map_items(func: Callable[[T], R], coll: Iterable[Iterable[T]]) -> Iterator[Iterator[R]]:
    """
    Lazy apply function `func` to all elements of the given iterable of iterables `coll`.
    *(Apply function for each element of each iterable from the given iterable.)*
    
    This function do not actuallyy iterate over the original iterable.
    To do so, use `functional.chaintools.apply()` and `functional.chaintools.apply_items()`.
    
    Examples:
        ```python
        c = chunks(range(10), 3)         # Something like: [ [0,1,2], [3,4,5], [6,7,8], [9] ]
        m = map_items(lambda x: x*x, c)  # Transform numbers into their squares
        print(list(map(list, m)))        # Simple print(list(m)) won't work here
        # [[0, 1, 4], [9, 16, 25], [36, 49, 64], [81]]
        ```
    
    Args:
        func: `Callable[[T], R]`. A function of signature `(T) -> R` to be applied for all elements within iterable of iterables.
        coll: `Iterable[Iterable[T]]`. Iterable of iterables which is going to be mapped.
    
    Returns:
        `Iterator[Iterator[R]]`. Returns an iterator of iterators of mapped elements.
    """
    
    return map(partial(map, func), coll)

def apply(func: Callable[[T], Any], coll: Iterable[T]):
    """
    General forEach method.
    Synchronously apply the given function `func` to all elements of iterable `coll`.
    Function results are ignored.
    
    Logically same as ```list(map(func, coll))```, but looks prettier.
    
    Args:
        func: `Callable[[T], Any]`. A function of signature `(T) -> R` to be applied for all elements within iterable.
        coll: `Iterable[T]`. Iterable which is going to be iterated.
    """
    
    for el in coll: func(el)

def apply_items(func: Callable[[T], Any], coll: Iterable[Iterable[T]]):
    """
    General forEachItem method.
    Synchronously apply the given function `func` to all elements of iterable of iterables `coll`.
    Function results are ignored.
    
    Logically same as ```list(map(list, map_items(func, coll)))```, but looks prettier.
    
    Examples:
        ```python
        c = chunks(range(10), 3)         # Chunks of size 3 of numbers from 0 to 9
        m = map_items(lambda x: x*x, c)  # Their squares
        apply_items(print, m)            # Print all these numbers one at a time
        ```
    
    Args:
        func: `Callable[[T], R]`. A function of signature `(T) -> R` to be applied for all elements within iterable of iterables.
        coll: `Iterable[Iterable[T]]`. Iterable of iterables which is going to be applied.
    """
    
    for it in coll:
        for el in it: func(el)

def invcall(func: Callable[[T], R], *args, **kwargs) -> R:
    """ Version of `functional.predef.call()` with inversed arguments order. """
    return call(*reversed(args), func, **kwargs)
def invmap(coll: Iterable[T], func: Callable[[T], R]) -> Iterator[R]:
    """ Version of `builtins.map()` with inversed arguments order. """
    return map(func, coll)
def invmap_items(coll: Iterable[Iterable[T]], func: Callable[[T], R]) -> Iterator[Iterator[R]]:
    """ Version of `functional.chaintools.map_items()` with inversed arguments order. """
    return map_items(func, coll)

@overload
def chain(el: T, func1: Callable[[T], R]) -> R:
    pass
@overload
def chain(el: T, func1: Callable[[T], T2], func2: Callable[[T2], R]) -> R:
    pass
@overload
def chain(el: T, func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], R]) -> R:
    pass
@overload
def chain(el: T, func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], R]) -> R:
    pass
@overload
def chain(el: T, func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], T5], func5: Callable[[T5], R]) -> R:
    pass
@overload
def chain(el: T, func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], T5], func5: Callable[[T5], T6], func6: Callable[[T6], R]) -> R:
    pass
def chain(el: T, *funcs: Callable[[T], R]) -> R:
    """
    Chain methods mapping for the given element.
    Just looks prettier than `func5(func4(func3(func2(func1(el)))))`.
    
    Examples:
        Calculate len of hex representation of the given number inputted as string:
        ```python
        chain('1234', int, hex, len) - 2
        # '1234' => 1234 => '0x4d2' => 5 => 3
        # Is the same as:
        len(hex(int('1234'))) - 2
        ```
    
    Args:
        el: A starting element for the sequence of maps.
        *funcs: Any number of functions to be chained. Left-most functions are applied first.
    
    Returns:
        Returns the result of the last function application.
    """
    
    # noinspection PyTypeChecker
    return reduce(invcall, funcs, el)

@overload
def chain_map(coll: Iterable[T], func1: Callable[[T], R]) -> Iterator[R]:
    pass
@overload
def chain_map(coll: Iterable[T], func1: Callable[[T], T2], func2: Callable[[T2], R]) -> Iterator[R]:
    pass
@overload
def chain_map(coll: Iterable[T], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], R]) -> Iterator[R]:
    pass
@overload
def chain_map(coll: Iterable[T], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], R]) -> Iterator[R]:
    pass
@overload
def chain_map(coll: Iterable[T], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], T5], func5: Callable[[T5], R]) -> Iterator[R]:
    pass
@overload
def chain_map(coll: Iterable[T], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], T5], func5: Callable[[T5], T6], func6: Callable[[T6], R]) -> Iterator[R]:
    pass
@overload
def chain_map(coll: Iterable[T], *funcs: List[Callable[[T2], R]]) -> Iterator[R]:
    pass
def chain_map(coll: Iterable[T], *funcs: Callable[[T2], R]) -> Iterator[R]:
    """
    Chain methods mapping for the given iterable.
    Just looks prettier than `map(lambda el: func5(func4(func3(func2(func1(el))))), coll)`
    or even evil devil `map(func5, map(func4, map(func3, map(func2, map(func1, el)))))`.
    
    Examples:
        Calculate len of hex representation of the given list of numbers inputted as string:
        ```python
        numbers = [ '123', '1234', '7', '18123' ]
        chain_map(numbers, int, hex, len, lambda l: l - 2)
        # '123'   => 123   => '0x7b'   => 4 => 2
        # '1234'  => 1234  => '0x4d2'  => 5 => 3
        # '7'     => 7     => '0x7'    => 3 => 1
        # '18123' => 18123 => '0x46cb' => 6 => 4
        # Result: [ 2, 3, 1, 4 ]
        
        # Is the same as:
        map(lambda x: len(hex(int(x))) - 2, numbers)
        map(lambda l: l - 2, map(len, map(hex, map(int, numbers))))
        ```
    
    Args:
        coll: `Iterable[T]`. A starting iterable for the sequence of maps.
        *funcs: Any number of functions to be chained. Left-most functions are applied first.
    
    Returns:
        Returns an iterable containing the results of the last function application.
    """
    
    # noinspection PyTypeChecker
    return reduce(invmap, funcs, coll)


@overload
def chain_map_items(coll: Iterable[Iterable[T]], func1: Callable[[T], R]) -> Iterator[Iterator[R]]:
    pass
@overload
def chain_map_items(coll: Iterable[Iterable[T]], func1: Callable[[T], T2], func2: Callable[[T2], R]) -> Iterator[Iterator[R]]:
    pass
@overload
def chain_map_items(coll: Iterable[Iterable[T]], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], R]) -> Iterator[Iterator[R]]:
    pass
@overload
def chain_map_items(coll: Iterable[Iterable[T]], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], R]) -> Iterator[Iterator[R]]:
    pass
@overload
def chain_map_items(coll: Iterable[Iterable[T]], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], T5], func5: Callable[[T5], R]) -> Iterator[Iterator[R]]:
    pass
@overload
def chain_map_items(coll: Iterable[Iterable[T]], func1: Callable[[T], T2], func2: Callable[[T2], T3], func3: Callable[[T3], T4], func4: Callable[[T4], T5], func5: Callable[[T5], T6], func6: Callable[[T6], R]) -> Iterator[Iterator[R]]:
    pass
@overload
def chain_map_items(coll: Iterable[Iterable[T]], *funcs: List[Callable[[T2], R]]) -> Iterator[Iterator[R]]:
    pass
def chain_map_items(coll: Iterable[Iterable[T]], *funcs: Callable[[T2], R]) -> Iterator[Iterator[R]]:
    """
    Chain methods mapping for the given iterable.
    The shortest analogues using the standard Python methods are:
    ```python
    gen = ((func5(func4(func3(func2(func1(el))))) for el in it) for it in coll)
    gen = map(lambda it: map(lambda el: func5(func4(func3(func2(func1(el))))), it), coll)
    # Compare with:
    gen = chain_map_items(coll, func1, func2, func3, func4, func5)
    ```
    Sometimes it may not be shorter, but takes a way less number for braces to play with.
    
    Note that unlike the analogues, this first applies func1 to all elements and only then proceeds to func2. On paper.
    In practice, it is all lazy computations.
    
    Examples:
        Calculate len of hex representation of the given list of lists of numbers inputted as string:
        ```python
        numbers = [ [ '1','2','3','4' ], [ '10', '21', '42', '63' ], [ '123', '234', '345', '999' ] ]
        chain_map_items(numbers, int, hex, len, lambda l: l - 2)
        # <computations same as for `functional.chaintools.chain_map()`>
        # Result: [ [1,1,1,1], [1,2,2,2], [2,2,3,3] ]
        
        # Is the same as:
        map_items(lambda x: len(hex(int(x))) - 2, numbers)
        ((len(hex(int(x))) - 2 for x in nums) for nums in numbers)
        ```
    
    Args:
        coll: `Iterable[T]`. A starting iterable for the sequence of maps.
        *funcs: Any number of functions to be chained. Left-most functions are applied first.
    
    Returns:
        Returns an iterable containing the results of the last function application.
    """
    
    # noinspection PyTypeChecker
    return reduce(invmap_items, funcs, coll)


__all__ = \
[
    'apply',
    'apply_items',
    'chain',
    'chain_map',
    'chain_map_items',
    'chunks',
    'invcall',
    'invmap',
    'invmap_items',
    'map_items',
]



# language=Markdown
"""
Decorations for Final classes.
Final classes are guarded from being inherited.

```python
from functional.final import final_class

@final_class
class MyFinalClass:
    def __init__(self, x):
        self.x = x

# The following would raise FinalInheritanceError
class ChildClass(MyFinalClass):
    def __init__(self, x = 5):
        super().__init__(x)
```

This is implemented by changing their
`__init_subclass__` method with the one throwing error.
However, any parent `__init_sublass__` are safe:

```python
from functional.final import final_class

class A:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.x = 4

@final_class
class B(A):
    pass

print(B.x) # Prints 4
```

"""

import sys
import warnings
from dataclasses import dataclass
from typing import *

from .typing_wrapper import *
from .util import PrettyException

C = TypeVar('C')

@dataclass(frozen=True)
class FinalInheritanceError(PrettyException, TypeError, Generic[C]):
    """ An exception that is thrown whenever final class is to be inherited from """
    
    cls: Type[C]
    """ A final class being inherited from """
    
    @property
    def message(self) -> str:
        return f"Cannot inherit from final class {self.cls.__qualname__!r}."

@dataclass(frozen=True)
class GenericFinalWarning(PrettyException, RuntimeWarning, Generic[C]):
    """ A warning that is thrown whenever generic final class is to be constructed on older versions of Python """
    
    cls: Type[C]
    """ A final class being constructed """
    
    @property
    def message(self) -> str:
        return f"Class {self.cls.__qualname__!r}: Attempting to construct generic final type. It could work not properly on Python {sys.version_info.major}.{sys.version_info.minor}.x"

@dataclass(frozen=True)
class FinalInheritanceWarning(RuntimeWarning, Generic[C]):
    """ A warning that is thrown whenever generic final class is to be inherited from on older versions of Python """
    
    cls: Type[C]
    """ A final class being inherited from """
    
    def __post_init__(self):
        super().__init__(self.message)
    
    @property
    def message(self) -> str:
        """ String representation of the error message """
        return f"Class {self.cls.__qualname__!r}: Attempting to inherit from generic final class. However, on Python {sys.version_info.major}.{sys.version_info.minor}.x this could not be handled"

def __init_subclass_exception__(cls, **kwargs):
    raise FinalInheritanceError(cls)

def __init_subclass_warning__(cls, **kwargs):
    warnings.warn(FinalInheritanceWarning(cls))

def final_class(cls: Type[C]) -> Type[C]:
    """
    A Decoration which makes the given class `cls` final.
    When inherited, the `FinalInheritanceError` will be raised.
    Also wraps it with `typing.final` decoration if exists, but some IDEs ignore it.
    
    Does not work properly for `typing.Generic` classes in Python 3.6.
    (Works fine for non-generic classes and/or on newer Pythons)
    
    Args:
        cls: A class to be made final.

    Returns:
        Returns the same class.
    
    Changes:
        Changed in version 0.1.2: Now throws warnings when used for Generic classes in Python 3.6
    """
    
    if (sys.version_info < (3, 7)):
        from typing_inspect import is_generic_type
        if (is_generic_type(cls)):
            cls_module, _, _ = getattr(cls, '__module__', '__main__').partition('.')
            this_module, _, _ = __name__.partition('.')
            if (this_module != cls_module):
                warnings.warn(GenericFinalWarning(cls))
            setattr(cls, '__init_subclass_warning__', classmethod(__init_subclass_warning__))
            cls.__init_subclass__ = cls.__init_subclass_warning__
            return final(cls)
    
    setattr(cls, '__init_subclass_exception__', classmethod(__init_subclass_exception__))
    cls.__init_subclass__ = cls.__init_subclass_exception__
    return final(cls)


__all__ = \
[
    # Own implementations
    'FinalInheritanceError',
    'FinalInheritanceWarning',
    'GenericFinalWarning',
    'final_class',
]



"""
A group of helper classes implementing filters.
Filters are function-like classes those can, well,
filter the given sequence for the given conditions.

The main difference between them and normal functions or lambdas
is the fact all classes in this file are frozen dataclasses and support hashing.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import *

__pdoc__ = dict()

T = TypeVar('T')
@dataclass(frozen=True)
class AbstractFilter(Generic[T], Callable[[T], bool], ABC):
    """
    A parent class for filters.
    Implements the core logic (yes) of filtering by calling the built-in `builtins.filter`
    which is a bit faster than lambdas or generator expressions.
    
    When inheriting from this class, make sure:
     * `check_element` method is implemented
     * Class is marked with `dataclass(frozen=True)` decoration
    
    Basically, all filter classes are same as the following:
    ```python
    filtered = filter(lambda el: my_func(el, param1, param2), seq)
    filtered = MyFilter(param1, param2).filter(seq)
    filtered = filter(MyFilter(param1, param2), seq)
    ```
    
    Usage:
        ```python
        @dataclass(frozen=True)
        class GEFilter(AbstractFilter[int]):
            than: int
            
            def check_element(self, el: int) -> bool:
                return el >= self.than
        
        lst = [ -1, 3, 8, 5, 0, -6, 7 ]
        for el in GEFilter(5).filter(lst):
            print(el)
        # Output:
        # 8 5 7
        ```
    """
    
    @abstractmethod
    def check_element(self, el: T) -> bool:
        """
        An abstract method that implements per-element filtering.
        Returns `True` if the given element matches condition.
        
        Args:
            el: `T`
            An element to check.
        
        Returns:
            `bool`.
        """
        pass
    
    def filter(self, seq: Iterable[T]) -> Iterator[T]:
        """
        Filters the given sequence via itself.
        
        Args:
            seq: `Iterable[T]`.
                A sequence to filter.
        
        Returns:
            `Iterator[T]`.
                A filtered sequence.
        """
        
        # noinspection PyArgumentList
        return filter(self.check_element, seq)
    
    @overload
    def __call__(self, el: T) -> bool:
        """
        Same as `AbstractFilter.check_element()`
        
        See Also:
            * `AbstractFilter.check_element()`
        """
        pass
    def __call__(self, *args, **kwargs) -> bool:
        return self.check_element(*args, **kwargs)


@dataclass(frozen=True)
class HasAttrFilter(AbstractFilter, Generic[T]):
    """
    An `AbstractFilter` implementation of Python `lambda el: hasattr(el, attr_name)`.
    Filters all elements those have the given attribute.
    """
    
    attr_name: str
    """ An attribute name to check. """
    
    def check_element(self, el: T) -> bool:
        return hasattr(el, self.attr_name)

@dataclass(frozen=True)
class _IsNotNoneFilter(AbstractFilter[Optional[T]], Generic[T]):
    """
    An `AbstractFilter` implementation of Python `lambda el: el is not None`.
    Filters-out all None elements.
    
    This class is a singleton.
    """
    
    def check_element(self, el: Optional[T]) -> bool:
        return el is not None
    
    def filter(self, seq: Iterable[Optional[T]]) -> Iterator[T]:
        pass

_IsNotNoneFilter.filter = AbstractFilter.filter
IsNotNoneFilter = _IsNotNoneFilter()
"""
An `AbstractFilter` implementation of Python `lambda el: el is not None`.
Filters-out all None elements.
"""


__all__ = \
[
    'AbstractFilter',
    'HasAttrFilter',
    'IsNotNoneFilter',
]
__pdoc__['_IsNotNoneFilter'] = True
__pdoc__['IsNotNoneFilter'] = True



# language=Markdown
"""
Predef Module -- predefined simple functions.
"""

from typing import *

T = TypeVar('T')
R = TypeVar('R')
def identity(x: T) -> T:
    """
    Identity Function -- returns its argument unchanged.
    A shorter alias for `lambda x: x`.
    
    Examples:
        Useful for flattening data:
        ```python
        def flatten(monad: Monad[K, Monad[K, T]]) -> Monad[K, T]:
            return monad.flat_map(identity)
        ```
    """
    
    return x

def call(func: Callable[[T], R], *args, **kwargs) -> R:
    """
    Calls the given function `func` with the given `*args` and `**kwargs`.
    A shorter alias for `lambda f, *args, **kwargs: f(*args, **kwargs)`.
    """
    return func(*args, **kwargs)

def star_call(func: Callable[[T], R], args, **kwargs) -> R:
    """
    Calls the given function `func` with the given `*args` and `**kwargs`.
    Applies starring to the `args`.
    A shorter alias for `lambda f, args, **kwargs: f(*args, **kwargs)`.
    """
    return func(*args, **kwargs)

def starmap_call(func: Callable[[T], R], kwargs, **extra_kwargs) -> R:
    """
    Calls the given function `func` with the given `**kwargs`.
    Applies star-mapping to the `kwargs`.
    A shorter alias for `lambda f, kwargs, **extra_kwargs: f(**kwargs, **extra_kwargs)`.
    """
    return func(**kwargs, **extra_kwargs)

__all__ = \
[
    'call',
    'identity',
    'star_call',
    'starmap_call',
]



# language=Markdown
"""
Class `AnyVal[T]` -- type-safe replacement for `typing.NewType('T')`.
"""

from abc import ABC
from typing import *

try:
    from dataclasses_json import global_config as DCJ_GLOBAL_CONFIG
    from dataclasses_json.core import Json
except ImportError:
    DCJ_SUPPORT_ENABLED = False
    DCJ_GLOBAL_CONFIG: Dict[type, Callable] = dict()
    Json = Union[dict, list, str, int, float, bool, None]
else:
    DCJ_SUPPORT_ENABLED = True

from .final import final_class
from .typing_wrapper import *

T = TypeVar('T')
C = TypeVar('C')

class AnyVal(ABC, Generic[T]):
    """
    Helper abstract class to make Scala-like AnyVal's.
    AnyVal is a dataclass-like class with the only field `value`,
    constructor, hash, representation, and equals, as well as encode/decode methods.
    
    If package `dataclasses_json` is installed,
    AnyVal subclasses are registered to have simple decoders and encoders.
    AnyVal subclasses are made to be final.
    
    Generally, works similar to `typing.NewType`, but the field `value` MUST be accepted explicitly.
    
    See Also:
        `functional.final.final_class`
    """
    
    __slots__ = ('value', )
    
    value: Final[T]
    """ An internal value of the `AnyVal` class """
    
    # region Constructors
    def __init__(self, value: T):
        self.value = value
    
    def __init_subclass__(cls, **kwargs):
        if (DCJ_SUPPORT_ENABLED):
            DCJ_GLOBAL_CONFIG.decoders[cls] = cls._decode
            DCJ_GLOBAL_CONFIG.encoders[cls] = cls._encode
        
        super().__init_subclass__(**kwargs)
        final_class(cls)
    # endregion
    
    # region Data Model Functions
    def __hash__(self):
        return hash(self.value)
    
    def __eq__(self, other):
        if (isinstance(other, AnyVal)):
            return type(self) == type(other) and self.value == other.value
        return False
    
    def __repr__(self):
        return f'{self.__class__.__name__}({self.value!r})'
    # endregion
    
    # region JSON Coders
    @classmethod
    def decode_value(cls, data: Json) -> T:
        """ Decode data of JSON-accepted type (like dict, str, etc...) and parse internal value from it. """
        return data
    def encode_value(self) -> Json:
        """ Encode internal value to the JSON-accepted type (like dict, str, etc...). """
        return self.value
    
    @classmethod
    def _decode(cls, data):
        return cls(cls.decode_value(data))
    def _encode(self):
        return self.encode_value()
    # endregion


__all__ = \
[
    'AnyVal',
]



# language=Markdown
"""
Class `FContainer[T]` -- an abstract container defining abstract mapping, iterative, and other methods.
"""

from abc import ABC, abstractmethod
from typing import *

from .monads import *

T = TypeVar('T')
R = TypeVar('R')
K = TypeVar('K', bound=Monad)
class FContainer(Monad[K, T], Functor[K, T], Generic[K, T], Container[T], ABC):
    """
    Abstract class representing an abstract container holding data of type `T`.
    It is also `Monad` and `Functor` over `T`.
    """
    
    __slots__ = tuple()
    
    empty: K
    """ A special class property containing the empty instance. """
    
    @property
    @abstractmethod
    def is_empty(self) -> bool:
        """ Returns True if the container is empty, False otherwise. """
        raise NotImplementedError
    
    @property
    @abstractmethod
    def non_empty(self) -> bool:
        """ Returns True if the container is non-empty, False otherwise. """
        raise NotImplementedError
    
    @abstractmethod
    def foreach(self, f: Callable[[T], Any]) -> None:
        """
        Apply the given function `f` to all values of the current container.
        The function results is ignored.
        
        This application can be either synchronous or lazy, depending on the implementation.
        
        Args:
            f: `Callable[[T], Any]`. A function of signature `(T) -> ...` that is applied.
        """
        
        raise NotImplementedError


__all__ = \
[
    'FContainer',
]



# language=Markdown
"""
Class `Option[T]` -- type-safe replacement for `typing.Optional('T')`.
"""

import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import *

from .containers import *
from .final import final_class
from .monads import *
from .predef import *
from .typing_wrapper import *
from .util import unmake_dataclass, PrettyException

__pdoc__ = dict()

DISABLE_WARNING: bool
""" Module global variable storing if `EmptySomeWarning` should be ignored or not. """

DISABLE_WARNING = False

@dataclass(frozen=True)
class EmptyOption(PrettyException, ValueError):
    """ An exception that is thrown whenever an empty `Option` is attempted to extract its value. """
    
    @property
    def message(self) -> str:
        return "Could not 'get' from an empty Option"

@dataclass(frozen=True)
class EmptySomeWarning(PrettyException, UserWarning):
    """ A warning that is thrown whenever Python's `None` is about to be put into the `Some` class (non-empty `Option`). """
    
    @property
    def message(self) -> str:
        return ' '.join \
        ([
            "'Some' Option with 'None' value detected.",
            "Usually this is an error.",
            "If it is not, you can suppress this warning by setting `functional.option.DISABLE_WARNING = True`.",
        ])

T = TypeVar('T')
R = TypeVar('R')
K = TypeVar('K', bound=Monad)
@unmake_dataclass
@dataclass(frozen=True, init=False, repr=False)
class Option(FContainer['Option[T]', T], Monad['Option[T]', T], Functor['Option[T]', T], Generic[T], ABC):
    """
    Represents optional values.
    Instances of Option are either an instance of `Some` or the object `Option.empty`.
    Options are generics of single type parameter.
    
    Examples:
        Scala-like constructor:
        ```python
        x = Some(4)      # Some(4)
        y = Option.empty # Option.empty
        z = none         # Option.empty
        ```
        
        Python-like constructor:
        ```python
        x = Option(4)    # Some(4)
        y = Option(None) # Option.empty
        ```
    """
    
    __slots__ = tuple()
    
    def __new__(cls, value: Optional[T], *, _ignore_selector: bool = False):
        if (_ignore_selector):
            return super().__new__(cls)
        elif (value is None):
            return OptionNone
        else:
            return Some(value)
    
    @classmethod
    def from_optional(cls: Type['Option[T]'], value: Optional[T]) -> 'Option[T]':
        """
        Constructs a new `Option[T]` from the given `typing.Optional[T]`.
        Returns `Option.empty` if the argument is Python `None`, `Some(arg)` otherwise.
        
        Args:
            value: `Optional[T]`. Either `None` or some value.
        
        Returns:
            `Option[T]`. Either `Option.empty` or `Some(...)`.
        """
        
        return Option(value)
    
    @classmethod
    def is_option(cls: Type['Option[T]'], value: Any) -> bool:
        """
        Class method which checks if the given value is an `Option`.
        Returns `True` if it is.
        
        Args:
            value: A potential `Option` to check.
        
        Returns:
            `bool`
        
        """
        return isinstance(value, Option)
    
    #region Properties
    @property
    def is_defined(self) -> bool:
        """ Returns True if the current Option is non-empty, False otherwise. """
        return not self.is_empty
    
    @property
    def non_empty(self) -> bool:
        """ Returns True if the current Option is non-empty, False otherwise. """
        return self.is_defined
    
    @property
    @abstractmethod
    def get(self) -> T:
        """
        Extracts the internal value of the current Option.
        Raises `EmptyOption` error if the Option is empty.
        
        Returns:
            `T`: Internal Option value. Potentially guarded that it is not `None`.
        """
        raise NotImplementedError
    
    @property
    def flatten(self) -> T:
        """
        Transforms `Option[Option[T]]` into `Option[T]`.
        Returns `Option.empty` if either the current Option, or its internal value is empty,
        and `Some(...)` otherwise.
        
        Returns:
            `Option[T]`: A flattened Option.
        """
        
        return self.flat_map(identity)
    
    @property
    @abstractmethod
    def as_optional(self) -> Optional[T]:
        """
        Transforms current `Option[T]` into Python `typing.Optional[T]`.
        Empty Options are transformed into None, other into the internal data.
        
        Returns:
            `typing.Optional[T]`. Either `None`, or some data.
        """
        
        raise NotImplementedError
    
    @abstractmethod
    def tuple_transform(self, num_items: int) -> Tuple['Option[T]', ...]:
        """
        Transforms an Option of Tuples into a Tuple of Options.
        
        Examples:
            ```python
            Some((1, 2, 3)).tuple_transform(3)  # (Some(1), Some(2), Some(3))
            Option.empty.tuple_transform(3)     # (Option.empty, Option.empty, Option.empty)
            ```
        
        Args:
            num_items: `int`. A number of items in the self-tuple.
                This argument is ignored for non-empty Options.
        
        Returns:
            `Tuple[Option[T], ...]`: A tuple of Options storing the internal data.
        """
        
        raise NotImplementedError
    #endregion
    
    #region Methods
    @abstractmethod
    def get_or_else(self, or_else: T) -> T:
        """
        Getter of the option value with a non-lazy fallback.
        Returns internal value for non-empty Options, and `or_else` otherwise.
        
        If you are interested in the lazy fallback, use `or` operator (see: Option.__bool__).
        
        Args:
            or_else: `T`. A non-lazy fallback.
        
        Returns:
            `T`: Either Option internal value, or the `or_else` argument.
        """
        
        raise NotImplementedError
    
    @abstractmethod
    def flat_map(self, f: Callable[[T], 'Option[R]']) -> 'Option[R]':
        """
        Maps the value of the current Option using the given function `f` into a new Option..
        Empty Options remain empty, non-empty Options may become non-empty.
        
        Unlike many other mapping functions, this is one synchronous (non-lazy).
        
        Args:
            f: `Callable[[T], Option[R]]`. A function of signature `(T) -> Option[R]` that is applied to the non-empty options.
        
        Returns:
            `Option[R]`: Option.empty, or the result of function `f`.
        """
        
        raise NotImplementedError
    
    def map(self, f: Callable[[T], R]) -> 'Option[R]':
        """
        Maps the value of the current Option using the given function `f`.
        Empty Options remain empty, non-empty Options remain non-empty.
        
        Unlike many other mapping functions, this is one synchronous (non-lazy).
        
        Args:
            f: `Callable[[T], R]`. A function of signature `(T) -> R` that is applied to the non-empty options.
        
        Returns:
            `Option[R]`: A new option potentially holding the result of mapping.
        """
        
        return self.flat_map(lambda x: Some(f(x)))
    
    def foreach(self, f: Callable[[T], Any]):
        """
        If the current Option is non-empty, apply function `f` to its internal value.
        Does nothing for empty Options.
        
        Same as `Option.map`, but (a) the mapping is guaranteed to be synchronously happen and (b) the result is ignored.
        
        Args:
            f: `Callable[[T], R]`. A function of signature `(T) -> R` that is applied to the non-empty options.
        """
        
        for v in self: f(v)
    #endregion
    
    #region Operators
    def __bool__(self) -> bool:
        return self.is_defined
    
    def __len__(self) -> int:
        return int(self.is_defined)
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    
    @abstractmethod
    def __contains__(self, item: T) -> bool:
        raise NotImplementedError
    
    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        raise NotImplementedError
    #endregion

@final
@final_class
@unmake_dataclass
@dataclass(frozen=True, repr=False)
class Some(Option[T], Generic[T]):
    """
    A class representing an instance of non-empty `Option`.
    
    Changes:
        Changed in version 0.2.0: Now throws warnings when `Some(None)` is called.
    """
    
    __slots__ = ('_value', )
    _value: T
    
    def __new__(cls, *args, **kwargs):
        kwargs['_ignore_selector'] = True
        return super().__new__(cls, *args, **kwargs)
    
    def __post_init__(self):
        if (self._value is None and not DISABLE_WARNING):
            warnings.warn(EmptySomeWarning())
    
    #region Properties
    @property
    def is_empty(self):
        return False
    
    @property
    def get(self) -> T:
        return self._value
    
    @property
    def as_optional(self) -> T:
        return self._value
    
    def tuple_transform(self, num_items: int) -> Tuple['Some[T]', ...]:
        return tuple(map(Some, self._value))
    # endregion
    
    # region Methods
    def get_or_else(self, or_else: T) -> T:
        return self._value
    
    def flat_map(self, f: Callable[[T], Option[R]]) -> Option[R]:
        return f(self._value)
    # endregion
    
    # region Operators
    def __repr__(self):
        return f"Some({self._value!r})"
    
    def __contains__(self, item: T) -> bool:
        return self._value == item
    
    def __iter__(self) -> Iterator[T]:
        yield self._value
    # endregion

@final
@final_class
@unmake_dataclass
@dataclass(frozen=True, repr=False, init=False)
class _OptionNone(Option[T], Generic[T]):
    """
    A class representing an instance of an empty `Option`.
    This class SHOULD have only one instance which is stored at `Option.empty`.
    Any other attempts to construct a new instance of this class are illegal.
    """
    
    __slots__ = tuple()
    
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, None, _ignore_selector=True)
    
    # region Properties
    @property
    def is_empty(self):
        return True
    
    @property
    def get(self) -> T:
        raise EmptyOption()
    
    @property
    def as_optional(self) -> None:
        return None
    
    def tuple_transform(self, num_items: int) -> Tuple['_OptionNone[T]', ...]:
        return tuple([ OptionNone ] * num_items)
    # endregion
    
    # region Methods
    def get_or_else(self, or_else: T) -> T:
        return or_else
    
    def flat_map(self, f: Callable[[T], Option[R]]) -> '_OptionNone[R]':
        # noinspection PyTypeChecker
        return none
    # endregion
    
    # region Operators
    def __str__(self):
        return 'None'
    
    def __repr__(self):
        return 'Option.empty'
    
    def __contains__(self, item: T) -> bool:
        return False
    
    def __iter__(self):
        return
        # noinspection PyUnreachableCode
        yield
    # endregion

T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
T6 = TypeVar('T6')

@overload
def tuple_transform(opt: Option[Tuple[T1]], num_items: int) -> Tuple[Option[T1]]:
    pass
@overload
def tuple_transform(opt: Option[Tuple[T1, T2]], num_items: int) -> Tuple[Option[T1], Option[T2]]:
    pass
@overload
def tuple_transform(opt: Option[Tuple[T1, T2, T3]], num_items: int) -> Tuple[Option[T1], Option[T2], Option[T3]]:
    pass
@overload
def tuple_transform(opt: Option[Tuple[T1, T2, T3, T4]], num_items: int) -> Tuple[Option[T1], Option[T2], Option[T3], Option[T4]]:
    pass
@overload
def tuple_transform(opt: Option[Tuple[T1, T2, T3, T4, T5]], num_items: int) -> Tuple[Option[T1], Option[T2], Option[T3], Option[T4], Option[T5]]:
    pass
@overload
def tuple_transform(opt: Option[Tuple[T1, T2, T3, T4, T5, T6]], num_items: int) -> Tuple[Option[T1], Option[T2], Option[T3], Option[T4], Option[T5], Option[T6]]:
    pass
def tuple_transform(opt: Option[Tuple[T, ...]], num_items: int) -> Tuple[Option[T], ...]:
    """ An alias for `Option.tuple_transform()`. """
    return opt.tuple_transform(num_items)

Option.empty = _OptionNone()
OptionNone = Option.empty
""" An alias for `Option.empty`. """
none = Option.empty
"""
An alias for `Option.empty`.
Pending deprecation since version 0.2.0.
Would be deprecated in version 0.3.0.
"""
is_option = Option.is_option
""" An alias for `Option.is_option`. """

def as_optional(o: Option[T]) -> Optional[T]:
    """
    An alias for `Option.as_optional`.
    
    Transforms given `Option[T]` into Python `typing.Optional[T]`.
    Empty Options are transformed into None, other into the internal data.
    
    Returns:
        `typing.Optional[T]`. Either `None`, or some data.
    """
    return o.as_optional


__all__ = \
[
    'EmptyOption',
    'EmptySomeWarning',
    'Option',
    'Some',
    'OptionNone',
    'none',
    'as_optional',
    'is_option',
    'tuple_transform',
]

__pdoc__['DISABLE_WARNING'] = True
__pdoc__['_OptionNone'] = True



# language=Markdown
"""
Typing Wrapper -- forward-compatibility for the Python `typing` module.

## Typing Wrapper

This module provides forwards-compatibility for the Python `typing` module.
It contains no actual members on the latest version of Python.
However, even these definitions are not available in the earlier Pythons,
the majority of IDEs' bundled type checkers process them as if they are.
"""

try: from typing import Final
except ImportError:
    from typing_extensions import Final

try: from typing import final
except ImportError:
    from typing_extensions import final



"""
.. include:: ../../README.md
"""

from collections import namedtuple

__title__ = 'functional-python'
__author__ = 'Peter Zaitcev / USSX Hares'
__license__ = 'BSD 2-clause'
__copyright__ = 'Copyright 2019-2022 Peter Zaitcev'
__version__ = '0.3.0'

VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
version_info = VersionInfo(*__version__.split('.'), releaselevel='alpha', serial=0)

from .final import *
from .predef import *

from .anyval import *
from .filters import *
from .option import *

# The following modules are only available by direct import:
#  * builtins_wrapper
#  * chaintools
#  * containers
#  * dcj_support
#  * monads
#  * typing_wrapper
#  * util

__all__ = \
[
    'version_info',
    '__title__',
    '__author__',
    '__license__',
    '__copyright__',
    '__version__',
    *final.__all__,
    *predef.__all__,
    *anyval.__all__,
    *filters.__all__,
    *option.__all__,
]

# Suppress documentation generation for the root module.
__pdoc__ = { k: False for k in __all__ if not k.startswith('__') }
del __pdoc__['version_info']
del __pdoc__['OptionNone']
del __pdoc__['IsNotNoneFilter']
del __pdoc__['none']



# language=Markdown
"""
Monadic classes `Functor[K,T]` and `Monad[K, T]` -- abstract monadic interfaces for Functors and Monads data types.
"""

from abc import ABC, abstractmethod
from typing import *

T = TypeVar('T')
R = TypeVar('R')
# noinspection PyTypeChecker
K = TypeVar('K', bound='Functor')
class Functor(Generic[K, T], ABC):
    """
    An abstract generic type of `Functor[K, T]`.
    Represents that the implementation of type `K` is generic over `T` and has methods supporting mapping.
    """
    
    __slots__ = tuple()
    
    @abstractmethod
    def map(self: K, f: Callable[[T], R]) -> K:
        """
        Maps the items in the given Monad and wraps result into a new Functor.
        
        Args:
            f: `Callable[[T], R]`. A function of signature `(T) -> R` which is applied to all elements in the current Functor.
        
        Returns:
            `K: Functor[K, R]`. A Functor containing results of application function `f` to the internal data.
        """
        
        raise NotImplementedError
del K

# noinspection PyTypeChecker
K = TypeVar('K', bound='Monad')
class Monad(Generic[K, T], ABC):
    """
    An abstract generic type of `Monad[K, T]`.
    Represents that the implementation of type `K` is generic over `T` and has methods supporting flat mapping.
    """
    
    __slots__ = tuple()
    
    @abstractmethod
    def flat_map(self, f: Callable[[T], K]) -> K:
        """
        Maps the items in the given Monad into a new Monad.
        
        Args:
            f: `Callable[[T], K]`. A function of signature `(T) -> K` where `T` is data type for the current Monad and `K` is the new Monad.
        
        Returns:
            `K: Monad[K, R]`. An aggregated result of application of function `f` to the internal data of the current Monad.
        """
        
        raise NotImplementedError
    
    @property
    @abstractmethod
    def flatten(self) -> T:
        """
        Flattens the current Monad -- transforms a Monad of Monads into a Monad (i.e., List of Lists into List).
        Signature: `Monad[K, Monad[K, T]] -> Monad[K, T]`.
        
        Returns:
            A flattened monad.
            Return value type is the same as the type of the current monad's internal data.
            Return value is an aggregated data of internal data.
        """
        
        raise NotImplementedError
del K

__all__ = \
[
    'Monad',
    'Functor',
]



# language=Markdown
"""
Utilities Module -- miscellaneous functions and classes used across the package.
"""

from abc import ABC, abstractmethod
from dataclasses import is_dataclass, dataclass
from typing import *

try:
    # noinspection PyProtectedMember
    from dataclasses import _FIELDS
except ImportError:
    _FIELDS = '__dataclass_fields__'

C = TypeVar('C')
def unmake_dataclass(cls: Type[C]) -> Type[C]:
    """
    Unregisters the given class `cls` from being a `dataclass`.
    It keeps all its DataClass properties (including dataclass nesting potential)
    but not its `__dataclass_fields__` which are the only criteria
    `dataclasses.is_dataclass()` decides if the argument is a dataclass.
    
    Actually modifies the existing class object rather than creates a new one.
    In other words, this statement is true: ```unmake_dataclass(cls) is cls```
    
    Args:
        cls: `Type[C]`. Class to be "undataclassed".
    
    Returns:
        `Type[C]`. Returns the same class.
    """
    
    if (not isinstance(cls, type)):
        raise TypeError(f"Dataclass type expected, got {type(cls).__name__!r}: {cls!r}")
    elif (not is_dataclass(cls)):
        raise TypeError("Unable to unmake class that is not a dataclass")
    
    delattr(cls, _FIELDS)
    assert not is_dataclass(cls), f"Assertion failed: {cls.__name__!r}"
    
    # noinspection PyTypeChecker
    return cls


@dataclass(frozen=True)
class PrettyException(Exception, ABC):
    """ Abstract class providing base to all exceptions in the package. """
    
    def __post_init__(self):
        super().__init__(self.message)
    
    @property
    @abstractmethod
    def message(self) -> str:
        """ String representation of the error message """
        raise NotImplementedError


__all__ = \
[
    'unmake_dataclass',
    'PrettyException',
]
